/***********************************************************************************************************************
 *
 * blueBill Mobile - Android - open source birding
 * Copyright (C) 2009-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluebill.tidalwave.it/mobile
 * SCM: https://java.net/hg/bluebill-mobile~android-src
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.observation.report;

import javax.annotation.Nonnull;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.TreeSet;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import it.tidalwave.util.AsException;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.geo.Coordinate;
import it.tidalwave.geo.Range;
import it.tidalwave.observation.Location;
import it.tidalwave.observation.Observation;
import it.tidalwave.observation.ObservationItem;
import it.tidalwave.observation.ObservationSet;
import it.tidalwave.bluebill.taxonomy.mobile.Taxon;
import it.tidalwave.semantic.document.Document;
import it.tidalwave.bluebill.mobile.preferences.GeneralPreferences;
import it.tidalwave.bluebill.mobile.preferences.TaxonomyPreferences;
import static it.tidalwave.role.HtmlRenderable.HtmlRenderable;
import static it.tidalwave.bluebill.mobile.observation.ui.ObservationItemRenderingOption.*;
import static it.tidalwave.bluebill.taxonomy.mobile.Taxon.Taxon;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 **********************************************************************************************************************/
public class KmlReportGenerator extends ReportGenerator
  {
    private static final double RADIUS_THRESHOLD_FOR_RANGE_CIRCLE = 10.0;

    @Nonnull
    private TaxonomyPreferences taxonomyPreferences  = Locator.find(TaxonomyPreferences.class);

    @Nonnull
    private final GeneralPreferences generalPreferences = Locator.find(GeneralPreferences.class);

    private boolean ignore = false;

    private final List<String> names = new ArrayList<String>();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public KmlReportGenerator() 
      {
        super("KML", Document.KML_MIME_TYPE);
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void visit (final @Nonnull ObservationSet observationSet)
      {
        buffer.append("<?xml version='1.0' encoding='UTF-8'?>\n"
                    + "<kml xmlns=\"http://www.opengis.net/kml/2.2\" xmlns:gx=\"http://www.google.com/kml/ext/2.2\">\n"
                    + "  <Document>\n"
                    + "    <name>blueBill Report</name>\n"
                    + "    <description>blueBill Report.</description>\n"
                    + "    <Style id=\"defaultStyles\">\n"
                    + "      <LineStyle>\n"
                    + "        <color>cc000000</color>\n"
                    + "      </LineStyle>\n"
                    + "      <PolyStyle>\n"
                    + "      </PolyStyle>\n"
                    + "    </Style>\n");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void postVisit (final @Nonnull ObservationSet observationSet)
      {
        buffer.append("  </Document>\n</kml>\n");
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void preVisit (final @Nonnull Observation observation)
      {
        try
          {
            final Location location = observation.getLocation();
            final Range range = location.as(Range); // can be not present, test before producing any output
            buffer.append("    <Placemark>\n");

            buffer.append("      <Point>\n");
            buffer.append("        <coordinates>").append(formatCoordinate(range.getCoordinate())).append("</coordinates>\n");
            buffer.append("      </Point>\n");

            final DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'hh:mm:ss'Z'");

            buffer.append("      <gx:TimeStamp>\n");
            buffer.append("        <when>").append(df.format(observation.getDate())).append("</when>\n");
            buffer.append("      </gx:TimeStamp>\n");

            names.clear();

            final Set<String> taxonNames = new TreeSet<String>();
            final Locale primaryLocale = taxonomyPreferences.getTaxonomyLocales().get(0);

            for (final ObservationItem item : observation.findObservationItems().results()) 
              {
                final Taxon taxon = item.getObservable().as(Taxon);
                try
                  {
                    taxonNames.add(taxon.getDisplayName(primaryLocale));
                  }
                catch (NotFoundException e)
                  {
                    taxonNames.add(taxon.getScientificName());
                  }
              }

            final StringBuilder name = new StringBuilder();
            String separator = "";
            int n = 0;

            for (final String taxonName : taxonNames)
              {
                if (++n >= 5)
                  {
                    name.append("...");
                    break;
                  }

                name.append(separator).append(taxonName);
                separator = ", ";
              }

            buffer.append(String.format("      <name>%s</name>\n", name.toString()));
            ignore = false;
          }
        catch (AsException e)
          {
            // Range not available, can't place on a map
            ignore = true;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void visit (final @Nonnull ObservationItem item)
      {
        if (!ignore)
          {
            names.add(item.as(HtmlRenderable).render(CARDINALITY, GENDER, MODE, NOTE_FULL));
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Override
    public void postVisit (final @Nonnull Observation observation)
      {
        if (!ignore)
          {
            Collections.sort(names);
            buffer.append("      <description>\n");
            buffer.append("        <![CDATA[\n");

            buffer.append(String.format("          <b>%s</b><br>\n",
                                        generalPreferences.formatDateAndTime(observation.getDate())));
            buffer.append(String.format("          <b>%s</b><br>\n",
                                        observation.getLocation().getDisplayName()));

            buffer.append("          <ul>\n");

            for (int i = 0; i < names.size(); i++)
              {
                buffer.append(String.format("            <li>%s</li>\n", names.get(i)));
              }

            buffer.append("          </ul>\n");
            buffer.append("        ]]>\n");
            buffer.append("      </description>\n");
            buffer.append("    </Placemark>\n");

            final Location location = observation.getLocation();
            final Range range = location.as(Range);

            if (range.getRadius() > RADIUS_THRESHOLD_FOR_RANGE_CIRCLE)
              {
                outputRangeCircle(range, buffer);
              }
          }    
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void outputRangeCircle (final @Nonnull Range range, final @Nonnull StringBuilder buffer)
      {
        final Coordinate coordinate = range.getCoordinate();
        final StringBuilder path = new StringBuilder();
        final int n = 20;

        for (double angle = 0; angle < 360; angle += 360 / n)
          {
            path.append(formatCoordinate(x(coordinate, range.getRadius(), angle))).append(" ");
          }

        buffer.append("    <Placemark>\n");
        buffer.append("      <styleUrl>#defaultStyles</styleUrl>\n");
        buffer.append("      <Style>\n");
        buffer.append("        <PolyStyle>\n");
        buffer.append("          <color>8000ffff</color>\n");
        buffer.append("        </PolyStyle>\n");
        buffer.append("      </Style>\n");
        buffer.append("      <Polygon>\n");
        buffer.append("        <outerBoundaryIs>\n");
        buffer.append("          <LinearRing>\n");
        buffer.append("            <coordinates>").append(path).append("</coordinates>\n");
        buffer.append("          </LinearRing>\n");
        buffer.append("        </outerBoundaryIs>\n");
        buffer.append("      </Polygon>\n");
        buffer.append("    </Placemark>\n");
        names.clear();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private static Coordinate x (final @Nonnull Coordinate from, final double range, final double bearing)
      {
        final double EARTH_RADIUS_EQUATOR = 6378140.0;

        final double b = Math.toRadians(bearing);
        final double lon = Math.toRadians(from.getLongitude());
        final double lat = Math.toRadians(from.getLatitude());
//        final double f = 1/298.257;
        final double e = 0.08181922;

        final double R = EARTH_RADIUS_EQUATOR * (1 - e * e) / Math.pow((1 - e * e * Math.pow(Math.sin(lat), 2)), 1.5);
        final double psi = range / R;
        final double phi = Math.PI / 2 - lat;
        final double arccos = Math.cos(psi) * Math.cos(phi) + Math.sin(psi) * Math.sin(phi) * Math.cos(b);
        final double latA = Math.toDegrees(Math.PI / 2 - Math.acos(arccos));

        final double arcsin = Math.sin(b) * Math.sin(psi) / Math.sin(phi);
        final double longA = Math.toDegrees(lon - Math.asin(arcsin));

        return new Coordinate(latA, longA, from.getAltitude());
      }

    @Nonnull
    private static String formatCoordinate (final @Nonnull Coordinate c)
      {
        return String.format(Locale.UK, "%f, %f, %f", c.getLongitude(), c.getLatitude(), c.getAltitude());
      }
  }
